"""
Event modeling.

"""
from microcosm_eventsource.models.alias import ColumnAlias  # noqa: F401
from microcosm_eventsource.models.base import BaseEvent  # noqa: F401
from microcosm_eventsource.models.meta import EventMeta  # noqa: F401
from microcosm_eventsource.models.rollup import RollUp  # noqa: F401



"""
Rolled up event store.

"""
from microcosm_postgres.context import SessionContext
from microcosm_postgres.errors import ModelNotFoundError
from sqlalchemy.orm import aliased
from sqlalchemy.orm.exc import NoResultFound

from microcosm_eventsource.func import ranked
from microcosm_eventsource.models.rollup import RollUp


class RollUpStore:

    def __init__(self, container_store, event_store, rollup=RollUp):
        self.container_store = container_store
        self.event_store = event_store
        self.rollup = rollup

    @property
    def model_class(self):
        return self.rollup

    @property
    def container_type(self):
        return self.container_store.model_class

    @property
    def event_type(self):
        return self.event_store.model_class

    def retrieve(self, identifier):
        """
        Retrieve a single rolled-up event.

        """
        container = self._retrieve_container(identifier)
        aggregate = self._aggregate()

        try:
            return self._to_model(
                aggregate,
                *self._filter(
                    self._rollup_query(
                        container,
                        aggregate,
                    ),
                    aggregate
                ).one(),
            )
        except NoResultFound as error:
            raise ModelNotFoundError(
                "{} not found".format(
                    self.container_type.__name__,
                ),
                error,
            )

    def count(self, **kwargs):
        """
        Query the number of possible rolled-up rows.

        Note that this count avoids joining across the event store; this logic works as long as
        every container row has at least one event row; we consider this a best practice.
        For the exact results - use exact_count.

        """
        return self.container_store.count(**kwargs)

    def exact_count(self, **kwargs):
        """
        Query the number of possible rolled-up rows.

        Note that this count joins across the event store - and costs more to calculate.

        """
        return self._search_query(**kwargs).count()

    def search(self, **kwargs):
        """
        Implement a rolled-up search of containers by their most recent event.

        """
        aggregate = self._aggregate(**kwargs)
        return [
            self._to_model(aggregate, *row)
            for row in self._search_query(aggregate, **kwargs).all()
        ]

    def search_first(self, **kwargs):
        results = self.search(**kwargs)
        if results:
            return results[0]
        return None

    def _search_query(self, aggregate=None, limit=None, offset=None, **kwargs):
        """
        Create the query for a rolled-up search of containers by their most recent event.

        Attempt to use the container object's store's filtering to limit the number of events
        that needs to be rolled up.

        """
        # SELECT
        #   <event.*>,
        #   <container.*>,
        #   rank...
        #   FROM (
        #     SELECT
        #       <event.*>,
        #       rank() OVER (
        #         PARTITION BY <event>.<container_id> ORDER BY <event>.clock DESC
        #       ) as rank
        #       ...
        #       FROM <event>
        #       JOIN (
        #         SELECT <container.*>
        #           FROM <container>
        #          WHERE <filter>
        #       ) as <container>
        #         ON <container>.id = <event>.<container_id>
        #      ORDER BY <order>
        #   )
        #   WHERE rank = 1
        #   LIMIT <limit> OFFSET <offset>

        # N.B. this method will handle the limit and offset instead of passing it down the
        # container store. This is to prevent unexpected results where the the number of
        # results returned from this method does not match the limit provided

        container = self._search_container(**kwargs)
        aggregate = aggregate or self._aggregate(**kwargs)
        query = self._filter(
            self._rollup_query(
                container,
                aggregate,
                **kwargs
            ),
            aggregate,
            **kwargs
        )

        if offset is not None:
            query = query.offset(offset)
        if limit is not None:
            query = query.limit(limit)

        return query

    def _retrieve_container(self, identifier):
        """
        Generate a subquery against the container type, filtering by identifier.

        """
        # SELECT * FROM <container> WHERE id = <identifier>
        return self._container_subquery(
            self.container_store._query().filter(
                self.container_type.id == identifier,
            )
        )

    def _search_container(self, **kwargs):
        """
        Generate a subquery against the container type, filtering by kwargs.

        """
        # SELECT * FROM <container> WHERE <filter>
        return self._container_subquery(
            self.container_store._filter(
                self.container_store._order_by(
                    self.container_store._query(),
                ),
                **kwargs
            )
        )

    def _container_subquery(self, query):
        """
        Wrap a container query so that it can be used in an aggregation.

        If the container class is polymorphic, then the operation reduces columns - in case
        of redundant names (e.g. joined table inheritance), otherwise operation explicitly maps
        the subquery to the table name so that ordering can be applied in `_rollup_query` without modifying
        the container store. This operation further wraps the subquery in an alias
        to the container type so that it can be referenced in `_rollup_query` as an entity.

        """
        if self._is_joined_polymorphic():
            subquery = query.subquery(
                reduce_columns=True,
                with_labels=True,
            )
        else:
            subquery = query.subquery(
                self.container_type.__tablename__,
            )

        return aliased(self.container_type, subquery)

    def _aggregate(self, **kwargs):
        """
        Emit a dictionary of window functions by which to aggregate events.

        By default, `_aggregate` performs a rank by clock values to pick the most recent row per logical time.

        """
        # rank() OVER (partition by <container_id> order by clock desc)
        # ... <possibly more>
        return dict(
            rank=ranked(self.event_type),
        )

    def _rollup_query(self, container, aggregate, **kwargs):
        """
        Query for events and aggregates that match the container subquery.

        We expect that the number of (filtered) container rows (e.g. via a LIMIT statement) to be much less than
        the number of total event rows. Note that if we were to query by events first and then by the container
        type, we might select many more rows.

        The query applies `from_self` so that aggregates can be legally used in the `_filter` WHERE clause.
        (Window functons are otherwise not allowed in WHERE because they are applied after other conditions.)

        """
        # SELECT *
        #   FROM (
        #     SELECT <event.*>, <container.*>, rank
        #       FROM <event_type>
        #       JOIN <container_type> ON <container_type>.id = <event_type>.container_id
        #      ORDER BY <order>
        #   )
        return self.container_store._order_by(
            self._query(
                container,
                aggregate,
            ),
            **kwargs
        ).from_self()

    def _query(self, container, aggregate):
        """
        Query events, containers, and aggregates together.

        """
        query = SessionContext.session.query(
            self.event_type,
            container,
        ).add_columns(
            *aggregate.values(),
        ).join(
            container,
            container.id == self.event_type.container_id,
        )

        if self._is_joined_polymorphic():
            return query.join(
                # extra join for reusing ordinary `_order_by` from container_store
                self.container_type,
                self.container_type.id == self.event_type.container_id,
            )

        return query

    def _filter(self, query, aggregate, **kwargs):
        """
        Filter by aggregates.

        By default, selects only events with the top rank (e.g. most recent clock).

        """
        return query.filter(
            aggregate["rank"] == 1,
        )

    def _to_model(self, aggregate, event, container, *args):
        keys = aggregate.keys()
        values = args[:len(keys)]
        return self.rollup(
            event,
            container,
            {
                key: value
                for (key, value) in zip(keys, values)
            },
            *args[len(keys):]
        )

    def _is_joined_polymorphic(self):
        if hasattr(self.container_type, "__mapper_args__"):
            return all([
                # joined polymorphic entity must have a polymorphic_identity set
                "polymorphic_identity" in self.container_type.__mapper_args__,
                # and there must exist foreign key on the identifier
                hasattr(self.container_type.id, "foreign_keys"),
            ])

        return False



"""
Custom database functions.

"""
from pkg_resources import resource_string

from microcosm_postgres.models import Model
from sqlalchemy import DDL, func
from sqlalchemy.event import listen
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql.expression import FunctionElement


def load_ddl(name, action):
    return resource_string("microcosm_eventsource", f"ddl/{name}.{action}.ddl").decode("utf-8")


# register custom DDL when calling `create_all` or `drop_all`
listen(
    Model.metadata,
    "after_create",
    DDL(
        load_ddl("array_sort_unique", "create") +
        load_ddl("proc_events_create", "create") +
        load_ddl("proc_events_delete", "create") +
        load_ddl("proc_event_type_delete", "create") +
        load_ddl("last_agg_sfunc", "create") +
        load_ddl("last_agg", "create") +
        load_ddl("proc_event_type_replace", "create")
    ),
)
listen(
    Model.metadata,
    "after_drop",
    DDL(
        load_ddl("array_sort_unique", "drop") +
        load_ddl("proc_events_create", "drop") +
        load_ddl("proc_events_delete", "drop") +
        load_ddl("proc_event_type_delete", "drop") +
        load_ddl("last_agg", "drop") +
        load_ddl("last_agg_sfunc", "drop") +
        load_ddl("proc_event_type_replace", "drop")
    ),
)


class last(FunctionElement):
    """
    Define a SQLAlchemy function that maps to a postgres function to select the last non-null value in window.

    """
    name = "last"

    @classmethod
    def of(cls, column, *filter_by):
        """
        Generate a window function over an event's column.

        """
        event = column.class_
        return cls(column).filter(*filter_by).over(
            order_by=event.clock.asc(),
            partition_by=event.container_id,
        )


@compiles(last)
def compile(element, compiler, **kwargs):
    """
    By default, `last` is not defined.

    """
    raise NotImplementedError("last is only defined for postgresql")


@compiles(last, "postgresql")
def compile_postgres(element, compiler, **kwargs):
    """
    Define `last` in terms of the custom functions.

    """
    return f"last_agg({compiler.process(element.clauses)})"


def ranked(event_type):
    return func.rank().over(
        order_by=event_type.clock.desc(),
        partition_by=event_type.container_id,
    )



"""
Event resources.

"""
from marshmallow import (
    Schema,
    ValidationError,
    fields,
    validates_schema,
)
from microcosm_flask.paging import PageSchema


class EventSchema(Schema):
    clock = fields.Integer(allow_none=True)
    createdTimestamp = fields.Float(
        attribute="created_timestamp",
        required=True,
        allow_none=True,
    )
    id = fields.UUID(
        required=True,
    )
    version = fields.Integer(
        required=True,
    )
    parentId = fields.UUID(
        attribute="parent_id",
        required=False,
        allow_none=True,
    )


class SearchEventSchema(PageSchema):
    clock = fields.Integer()
    min_clock = fields.Integer()
    max_clock = fields.Integer()
    parent_id = fields.UUID()
    sort_by_clock = fields.Boolean()
    sort_clock_in_ascending_order = fields.Boolean()
    version = fields.Integer()

    @validates_schema
    def validate(self, obj, **kwargs):
        if obj.get("sort_clock_in_ascending_order") and not obj.get("sort_by_clock"):
            raise ValidationError(
                "sort_by_clock must be set if sort_clock_in_ascending_order is set",
                field_names=[
                    "sort_by_clock",
                    "sort_clock_in_ascending_order",
                ],
            )



"""
Column aliasing.

"""


class ColumnAlias:
    """
    Descriptor to reference a column by a well known alias.

    """
    def __init__(self, name):
        self.name = name

    def __get__(self, cls, owner):
        return getattr(cls or owner, self.name)

    def __set__(self, cls, value):
        return setattr(cls, self.name, value)



"""
Event base class.

"""


class BaseEvent:

    def is_similar_to(self, other):
        """
        Are two events similar enough to activate upserting?

        """
        return all((
            self.event_type == other.event_type,
            self.parent_id == other.parent_id,
            self.container_id == other.container_id
        ))



"""
Event type enumerations.

As events are processed, they accumulate a `state` as a list of event types, using
a pluggable rule system:

 -  The state can be the current event: `current()`
 -  The state can be a union of the current event and the current state: `union()`
 -  The state can be another state: `alias()`

State machine transitions are defined using a DNF-compatible mini-language that processes
this accumulated `state` and the current event:

 -  An event can legally follow another another event: `event(name)`
 -  An event can legally follow a disjunction of conditions: `any_of(...)`
 -  An event can legally follow a conjunction of conditions: `all_of(...)`
 -  An event can legally follow a negation of a conditon: `but_not(...)`

"""
from enum import Enum
from itertools import chain

from microcosm_eventsource.accumulation import current
from microcosm_eventsource.errors import IllegalInitialStateError, IllegalStateTransitionError
from microcosm_eventsource.transitioning import nothing


class EventTypeInfo:
    """
    Event type meta data.

    """
    def __init__(self, follows=None, accumulate=None, restarting=False, requires=(), auto_transition=False):
        """
        :param follows:         an instance of the event mini-grammar
        :param accumulate:      whether the event type should accumulating state
        :param restarting:      whether the event type may restart a (new) version
        :param requires:        validate type-specific required fields
        :param auto_transition:  whether this event should be autogenereated when possible

        """
        self.follows = follows if follows is not None else nothing()
        self.accumulate = accumulate if accumulate is not None else current()
        self.restarting = restarting
        self.requires = requires
        self.auto_transition = auto_transition


class EventType(Enum):
    """
    Based event type enum.

    """
    @property
    def is_initial(self):
        """
        Can this event type be used for an initial event?

        """
        return not bool(self.value.follows)

    @property
    def is_accumulating(self):
        """
        Does this event type accumulate state?

        """
        return self.value.accumulating

    @property
    def is_restarting(self):
        """
        Does this event type restart a new version?

        """
        return self.value.restarting

    @property
    def is_auto_transition(self):
        """
        Does this event type would be autogenerated if possible?

        """
        return self.value.auto_transition

    @classmethod
    def auto_transition_events(cls):
        """
        Which events are auto-transition events?

        """
        return [event_type for event_type in cls if event_type.is_auto_transition]

    @classmethod
    def required_column_names(cls):
        return {
            column_name
            for event_type in cls
            for column_name in event_type.value.requires
        }

    @classmethod
    def requires(cls, column_name):
        return [
            event_type
            for event_type in cls
            if column_name in event_type.value.requires
        ]

    def may_transition(self, state):
        """
        Can this event type transition from the given state?

        :param state: a set of event types

        """
        return self.value.follows(self.__class__, state)

    @classmethod
    def available_transitions(cls, state):
        """
        To which events the given state may transition?

        :param state: a set of event types

        """
        return [event_type for event_type in cls if event_type.may_transition(state)]

    def validate_transition(self, state):
        """
        Asssert that a transition is legal from the given state.

        :param state: a set of event types

        """
        if self.may_transition(state):
            return

        if not state:
            # event may not be initial
            raise IllegalInitialStateError("Event type '{}' may not be initial".format(
                self.name,
            ))

        # event may not follow previous
        raise IllegalStateTransitionError("Event type '{}' may not follow [{}]".format(
            self.name,
            ", ".join(event_type.name for event_type in state),
        ))

    def accumulate_state(self, state):
        """
        Accumulate state.

        """
        return self.value.accumulate(state, self)

    def next_version(self, version):
        """
        Compute next version.

        """
        if version is None:
            return 1
        if self.is_restarting:
            return version + 1
        return version

    @classmethod
    def initial_states(cls):
        """
        Return a generator of all initial states.

        """
        for event_type in cls.available_transitions({}):
            yield frozenset(event_type.accumulate_state({}))

    @classmethod
    def all_states(cls):
        """
        Return a generator of all allowed states.

        """
        # Prefer to use frozenset and not set to represent states in this context
        # Hashable frozenset allows us to easily avoid repeating the same state
        next_states = list(cls.initial_states())
        known_states = set(next_states)

        while True:
            if not next_states:
                return
            current_states = next_states
            next_states = []
            for state in current_states:
                yield state
                for event_type in cls.available_transitions(state):
                    new_state = frozenset(event_type.accumulate_state(state))
                    if new_state in known_states:
                        continue
                    next_states.append(new_state)
                    known_states.add(new_state)

    @classmethod
    def all_states_and_events(cls):
        """
        Return a generator of all allowed (state, event_type) combinations
        Note: it can return the same state or event twice (but all uniqu)

        """
        next_states = [frozenset()]
        known_states = set(next_states)
        known_states_and_events = set()

        while True:
            if not next_states:
                return
            current_states = next_states
            next_states = []
            for state in current_states:
                for event_type in cls.available_transitions(state):
                    new_state = frozenset(event_type.accumulate_state(state))
                    if (new_state, event_type) in known_states_and_events:
                        continue
                    known_states_and_events.add((new_state, event_type))
                    yield new_state, event_type
                    if new_state in known_states:
                        continue
                    next_states.append(new_state)
                    known_states.add(new_state)

    @classmethod
    def all_transitions(cls, states=None):
        """
        Return a generator of all allowed transitions as a tuple of (initial state, new state, event type).

        :param state: a list states to check, If None - all allowed states

        """
        if states is None:
            states = cls.all_states()
        for state in states:
            for event_type in cls.available_transitions(state):
                new_state = frozenset(event_type.accumulate_state(state))
                yield (state, new_state, event_type)

    @classmethod
    def assert_only_valid_transitions(cls):
        """
        Check that the state machine has only valid transitions:
        * only one (or zero) auto-transition event can follow any state. No other event may follow.

        """
        for state in cls.all_states():
            auto_transition_events = [
                event_type for event_type in cls.auto_transition_events()
                if event_type.may_transition(state)
            ]
            if auto_transition_events and len(cls.available_transitions(state)) > 1:
                raise AssertionError("Only one auto-transition event can follow any state")

    def __lt__(self, other):
        return self.name < other.name

    def __str__(self):
        return self.name


def event_info(*args, **kwargs):
    return EventTypeInfo(*args, **kwargs)


def EventTypeUnion(name, *event_types):
    """
    Create a new `EventType` as a union of other enums.

    """
    return EventType(name, [(item.name, item.value) for item in chain(*event_types)])



"""
Event routes

"""
from microcosm_flask.conventions.base import EndpointDefinition
from microcosm_flask.conventions.crud import configure_crud
from microcosm_flask.operations import Operation
from microcosm_postgres.context import transactional
from microcosm_pubsub.producer import deferred_batch


def configure_event_crud(
    graph,
    controller,
    event_schema,
    new_event_schema,
    search_event_schema,
    use_deferred_batch=False,
):
    if use_deferred_batch:
        create_func = transactional(deferred_batch(controller)(controller.create))
    else:
        create_func = transactional(controller.create)

    mappings = {
        Operation.Create: EndpointDefinition(
            func=create_func,
            request_schema=new_event_schema,
            response_schema=event_schema,
        ),
        Operation.Delete: EndpointDefinition(
            func=transactional(controller.delete),
        ),
        Operation.Replace: EndpointDefinition(
            func=transactional(controller.replace),
            request_schema=new_event_schema,
            response_schema=event_schema,
        ),
        Operation.Retrieve: EndpointDefinition(
            func=controller.retrieve,
            response_schema=event_schema,
        ),
        Operation.Search: EndpointDefinition(
            func=controller.search,
            request_schema=search_event_schema,
            response_schema=event_schema,
        ),
    }
    configure_crud(graph, controller.ns, mappings)
    return controller.ns



"""
RolledUp event model.

"""


class RollUp:
    """
    Wrap a container and an event into rolled up state.

    The expectation is that attribute access will primarily refer to the container
    object and that specific event attributes will be exposed via @property syntax.

    """
    def __init__(self, event, container, aggregate, *args):
        """
        Roll ups are constructed with a single row from a roll up store query.

        By default, every row will contain:

        :param event: the most recent event in the roll up
        :param container: the container of the rolled up events
        :param aggregate: any other aggregates (window functions) across rolled up events

        """
        self._event = event
        self._container = container
        for key, value in aggregate.items():
            setattr(self, f"_{key}", value)

    @property
    def latest_event_type(self):
        """
        Expose the most recent event type.

        """
        return self._event.event_type

    def __getattr__(self, key):
        """
        By default, resolve attributes to the container.

        """
        return getattr(self._container, key)
